#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# classes for colors
#

import tree

#
# a seemingly good way to convert from string to integer
# for a couple of common ways of expressing colors ...
#

def color_str_to_int(str):
    if str.startswith('0x'):
        _val = int(str[2:], 16)
    elif str.startswith('#'):
        _val = int(str[1:], 16)
    else:
        _val = int(str)
    return _val

class Color(tree.TreeItem):
    """An object representing an RGB color.

The class purpose is self evident.

A Color object has three attributes:

r: Red (0-255)
g: Green (0-255)
b: Blue (0-255)

There is no alpha-channel attribute (yet)...

A Color object has the following methods:

getRed(): Get the Red value in the Color.
getBlue(): Get the Blue value in the Color.
getGreen(): Get the Green value in the Color.
clone(): Return an identical copy of a Color.

Once a color object is created, the values it
contains may not be changed.
    """
    def __init__(self, r=None, g=None, b=None):
        """Initialize a Color object.

There are several ways to create a color object:

Color(r,g,b) => r, g, and b values are all integers 0 <= value <= 255
Color(0xxxxx) => A hexidecimal value - it must begin with 0x. The color
                 is extracted from the value:
                 r = (0xff0000 & value) >> 16
                 g = (0xff00 & value) >> 8
                 b = (0xff & value)
Color('#hexvalue') => A string prefixed with '#', with the remaining
                      characters representing a hexidecimal value.
Color() => A default color with r = g = b = 255.
        """
        _r = r
        _g = g
        _b = b
        if isinstance(_r, str):
            val = color_str_to_int(_r)
            _r = (0xff0000 & val) >> 16
            _g = (0xff00 & val) >> 8
            _b = (0xff & val)
        elif isinstance(_r, int):
            if _g is None and _b is None:
                _r = (0xff0000 & r) >> 16
                _g = (0xff00 & r) >> 8
                _b = (0xff & r)
            elif _g is not None and _b is not None:
                if not isinstance(_g, int):
                    _g = int(g)
                if not isinstance(_b, int):
                    _b = int(b)
                if _r < 0 or _r > 255:
                    raise ValueError, "Red value must be between 0 and 255: %d" % _r
                if _g < 0 or _g > 255:
                    raise ValueError, "Green value must be between 0 and 255: %d" % _g
                if _b < 0 or _b > 255:
                    raise ValueError, "Blue value must be between 0 and 255: %d" % _b
            else:
                raise SyntaxError, "Invalid call to Color()."
        elif _r is None and _g is None and _b is None:
            _r = 255
            _g = 255
            _b = 255
        else:
            raise SyntaxError, "Invalid call to Color()."
        tree.TreeItem.__init__(self)
        self.__r = _r
        self.__g = _g
        self.__b = _b

    def __cmp__(self, obj):
        """Compare two Color objects.

The comparison is done based on the RGB values.

red value of C1 < red value of C2 ==> return -1
red value of C1 > red value of C2 ==> return 1

Then green values are compared, then blue. If
all values are equal, return 0.
        """
        if not isinstance(obj, Color):
            raise TypeError, "Invalid object for color comparison: " + `obj`
        return cmp(self.getColors(), obj.getColors())

    def __hash__(self):
        """Return a hash value for the Color object.

Providing this method means that Color objects can be used
as keys in dictionaries.
        """
        return (self.__r << 16) + (self.__g << 8) + self.__b
        
    def __repr__(self):
        return "Color(%d,%d,%d)" % (self.__r, self.__g, self.__b)
    
    def __str__(self):
        return "#%2.2x%2.2x%2.2x" % (self.__r, self.__g, self.__b)

    def getColors(self):
        """Return a three-item tuple with the values comprising this color.

getColors()        
        """
        return (self.__r, self.__g, self.__b)

    def getRed(self):
        """Return the red value of the color.

getRed()        
        """
        return self.__r

    r = property(getRed, None, None, "Red value of the color.")

    def getGreen(self):
        """Return the green value of the color.

getGreen()        
        """
        return self.__g

    g = property(getGreen, None, None, "Green value of the color.")

    def getBlue(self):
        """Return the blue value of the color.

getBlue()        
        """
        return self.__b

    b = property(getBlue, None, None, "Blue value of the color.")

    def clone(self):
        """Return a new Color object with the same color values.

clone()
        """
        return Color(self.__r, self.__g, self.__b)
