#
# Copyright (c) 2002, Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# classes for graphic objects
#

import style
import color
import linetype

class GraphicObject(object):
    """A class representing an object that is drawn.

This class is meant to be a base class for things like
line segments, circles, etc. The class has one attribute
that cannot be None:

style: The Style object

There are three other attributes that override values in
the style attribute:

color: A Color object
thickness: A positive float value
linetype: A Linetype object

A GraphicObject object has the following methods:

{get/set}Style(): Get/Set the Style of a GraphicObject.
{get/set}Color(): Get/Set the Color of a GraphicObject.
{get/set}Thickness(): Get/Set the thickness of a GraphicObject.
{get/set}Linetype(): Get/Set the Linetype of a GraphicObject.
inRegion(): Returns whether or not the GraphicObject is within a bounded area.
    """
    def __init__(self, st, lt=None, col=None, t=None):
        """Initialize a GraphicObject.

GraphicObject(st [, lt, col, t])

st: A style object - This argument cannot be None

Optional arguments:

lt: A Linetype object that overrides the value in the Style
col: A Color object that overrides the value in the Style
t: A positive float that overrides the value in the Style
        """
        _s = st
        if _s is None:
            raise ValueError, "All GraphicObjects must have a style."
        if not isinstance(_s, style.Style):
            _s = style.Style(st)
        _l = lt
        if _l is None:
            _l = _s.getLinetype()
        else:
            if not isinstance(_l, linetype.Linetype):
                raise TypeError, "Invalid linetype: " + `_l`
        _c = col
        if _c is None:
            _c = _s.getColor()
        else:
            if not isinstance(_c, color.Color):
                _c = color.Color(col)
        _t = t
        if _t is None:
            _t = _s.getThickness()
        else:
            if not isinstance(_t, float):
                _t = float(t)
        if _t < 0.0:
            raise ValueError, "Thickness must be greater than 0: " + `_t`
        self.__style = _s
        self.__color = _c
        self.__thickness = _t
        self.__linetype = _l
        
    def getStyle(self):
        """Get the Style of the GraphicObject.

getStyle()

Returns the current Style of an object. The values in the
style may be overriden if any of the color, linetype, or
thickness attributes are not None.
        """
        return self.__style

    def setStyle(self, s):
        """Set the Style of the Object.

setStyle(s)

Setting the style of a GraphicObject unsets any overrides for
the object.
        """
        _s = s
        if _s is None:
            raise ValueError, "All GraphicObjects must have a style."
        if not isinstance(_s, style.Style):
            _s = style.Style(s)
        self.__style = _s
        self.__color = _s.getColor()
        self.__thickness = _s.getThickness()
        self.__linetype = _s.getLinetype()

    style = property(getStyle, setStyle, None, "Object Style.")

    def getColor(self):
        """Get the Color of the Object.

getColor()
        """
        return self.__color

    def setColor(self, c):
        """Set the color of the object.

setColor(c)

Setting the color overrides the value in the Style. Setting
the color to None restores the value in the Style.
        """
        _c = c
        if _c:
            if not isinstance(_c, color.Color):
                _c = color.Color(c)
        else:
            _c = self.__style.getColor()
        self.__color = _c
        
    color = property(getColor, setColor, None, "Object color.")
    
    def getThickness(self):
        """Get the thickness of the GraphicObject.

getThickness()
        """
        return self.__thickness

    def setThickness(self, t):
        """Set the thickness of a object.

setThickness(t)

Setting the thickness overrides the value in the Style. Setting
the thickness to None restores the value in the Style.
        """
        _t = t
        if _t:
            if not isinstance(_t, float):
                _t = float(t)
            if _t < 0.0:
                raise ValueError, "Thickness must be greater than 0: " + `_t`
        else:
            _t = self.__style.getThickness()
        self.__thickness = _t

    thickness = property(getThickness, setThickness, None, "Object thickness.")
    
    def getLinetype(self):
        """Get the Linetype of the object.

getLinetype()
        """
        return self.__linetype

    def setLinetype(self, lt):
        """Set the Linetype of the GraphicObject.

setLinetype(lt)

Setting the Linetype overrides the value in the Sytle. Setting
the Linetype to None restores the value in the Style.
        """
        _l = lt
        if _l:
            if not isinstance(_l, linetype.Linetype):
                raise TypeError, "Invalid linetype: " + `_l`
        else:
            _l = self.__style.getLinetype()
        self.__linetype = _l

    linetype = property(getLinetype, setLinetype, None, "Object Linetype.")

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not an object falls within a boundary.

isRegion(xmin, ymin, xmax, ymax[, fully])

The first four arguments define the boundary. The optional
fifth argument "fully" indicates whether or not the object
must be completely contained within the region or just pass
through it.

This method should be overriden in classes derived from GraphicObjects.
        """
        return False
