#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# classes for line segments
#

import graphicobject
import tolerance
import style
import linetype
import color
import point
import tree
import maptree
import util

class _Segment(object):
    """A class representing a line segment.

A _Segment has two attributes.

p1: A _Point object representing the first end point
p2: A _Point object representing the second end point

A _Segment has the following methods:

getEndpoints(): Return the two endpoints of the _Segment.
{get/set}P1: Get/Set the _Segment first endpoint.
{get/set}P2: Get/Set the _Segment second endpoint.
move(): Move the _Segment.
length(): Get the _Segment length.
mapPoint(): Test if a point is within some distance to a _Segment.
inRegion(): Test if the _Segment is visible in some area.
clone(): Make an identical copy of a _Segment.
    """

    # member functions
    
    def __init__(self, p1, p2):
        """Instantiate a _Segment object.

_Segment(p1, p2)

p1: _Segment first endpoint - may be a _Point or a two-item tuple of floats
p2: _Segment second endpoint - may be a _Point or a two-item tuple of floats
        """
        _p1 = p1
        if not isinstance(_p1, point._Point):
            _p1 = point._Point(p1)
        _p2 = p2
        if not isinstance(_p2, point._Point):
            _p2 = point._Point(p2)
        if _p1 == _p2:
            raise ValueError, "_Segment objects cannot have equivalent endpoints."
        self.__p1 = _p1
        self.__p2 = _p2
        
    def __str__(self):
        return "Segment: %s to %s" % (self.__p1, self.__p2)

    def __cmp__(self, obj):
        """Compare two Segments or a Segment and a Point.

Segment comparison works as follows:
1) Get endpoints of both segments S1 and S2
2) Get minima and maxima values for the "x" and "y" coordinates

The return value is determined as follows:

minimum "x" of S1 < minimum "x" of S2 => return -1
maximum "x" of S1 > maximum "x" of S2 => return 1
minimum "y" of S1 < minimum "y" of S2 => return -1
maximum "y" of S1 > maximum "y" of S2 => return 1

Endpoint comparison now is needed to determine if
the Segments are equal.

if S1.p1.x == S2.p1.x and S1.p1.y == S2.p2.y or
   S1.p1.x == S2.p2.x and s1.p1.y == S2.p2.y ==> return 0
   
If the the above is not true, then the "x" coordinates of
the first endpoint of each segment are compared.

minimum "x" of S1E1 < minimum "x" of S2E1 => return -1
maximum "x" of S1E1 > maximum "x" of S2E1 => return 1

Segment/Point comparison works as follows:
1) Get endpoints of S and coordinates of P
2) Calculate minima and maxima of segment endpoints

The function returns the following

1: P.x < S.xmin
0: S.xmin <= P.x <= S.xmax and S.ymin <= P.y <= S.ymax
-1: Any other condition
        """
        if not isinstance(obj, (point._Point, _Segment)):
            raise TypeError, "Invalid object for Segment comparison: " + `obj`
        _sx1, _sy1 = self.__p1.getCoords()
        _sx2, _sy2 = self.__p2.getCoords()
        _sxmin = min(_sx1, _sx2)
        if isinstance(obj, _Segment):
            _p1, _p2 = obj.getEndpoints()
            _x1, _y1 = _p1.getCoords()
            _x2, _y2 = _p2.getCoords()
            _xmin = min(_x1, _x2)
            # print "sxmin = %g; xmin = %g" % (_sxmin, _xmin)
            if abs(_sxmin - _xmin) < 1e-10:
                _symin = min(_sy1, _sy2)
                _ymin = min(_y1, _y2)
                # print "symin = %g; ymin = %g" % (_symin, _ymin)
                if abs(_symin - _ymin) < 1e-10:
                    _sxmax = max(_sx1, _sx2)
                    _xmax = max(_x1, _x2)
                    # print "sxmax = %g; xmax = %g" % (_sxmax, _xmax)
                    if abs(_sxmax - _xmax) < 1e-10:
                        _symax = max(_sy1, _sy2)
                        _ymax = max(_y1, _y2)
                        # print "symax = %g; ymax = %g" % (_symax, _ymax)
                        if abs(_symax - _ymax) < 1e-10:
                            _x1diff = abs(_sx1 - _x1)
                            _y1diff = abs(_sy1 - _y1)
                            _x2diff = abs(_sx1 - _x2)
                            _y2diff = abs(_sy1 - _y2)
                            if ((_x1diff < 1e-10 and _y1diff < 1e-10) or
                                (_x2diff < 1e-10 and _y2diff < 1e-10)):
                                # print "common endpoints ..."
                                # print "self: " + str(self)
                                # print "obj: " + str(obj)
                                _val = 0
                            elif _sx1 > _x1:
                                _val = 1
                            else:
                                _val = -1
                        elif _symax > _ymax:
                            _val = 1
                        else:
                            _val = -1
                    elif _sxmax > _xmax:
                        _val = 1
                    else:
                        _val = -1
                elif _symin > _ymin:
                    _val = 1
                else:
                    _val = -1
            elif _sxmin > _xmin:
                _val = 1
            else:
                _val = -1
        else:
            _x, _y = obj.getCoords()
            if _x < _sxmin - 1e-10:
                _val = 1
            else:
                _val = -1
                _symin = min(_sy1, _sy2) - 1e-10
                _sxmax = max(_sx1, _sx2) + 1e-10
                _symax = max(_sy1, _sy2) + 1e-10
                if _x < _sxmax and _symin < _y < _symax:
                    _val = 0
        return _val
        
    def getEndpoints(self):
        """Get the endpoints of the Segment.

getEndpoints()

This function returns a tuple containing the two _Point objects
that are the endpoints of the segment.
        """
        return (self.__p1, self.__p2)

    def getP1(self):
        """Return the first endpoint _Point of the _Segment.

getP1()        
        """
        return self.__p1

    def setP1(self, p):
        """Set the first endpoint _Point of the _Segment.

setP1(p)
        """
        if not isinstance(p, point._Point):
            raise TypeError, "Invalid _Point for p1 endpoint: " + `p`
        self.__p1 = p
    
    p1 = property(getP1, setP1, None, "First endpoint of the Segment.")

    def getP2(self):
        """Return the second endpoint _Point of the _Segment.

getP2()        
        """
        return self.__p2

    def setP2(self, p):
        """Set the second endpoint _Point of the _Segment.

setP2(p)
        """
        if not isinstance(p, point._Point):
            raise TypeError, "Invalid Point for p2 endpoint: " + `p`
        self.__p2 = p

    p2 = property(getP2, setP2, None, "Second endpoint of the Segment.")

    def move(self, dx, dy):
        """Move a _Segment.

move(dx, dy)

The first argument gives the "x" coordinate displacement,
and the second gives the "y" coordinate displacement. Both
values should be floats.
        """
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        self.__p1.move(_dx, _dy)
        self.__p2.move(_dx, _dy)

    def length(self):
        """Return the length of the Segment.

length(s)
        """
        return self.__p1 - self.__p2

    def mapPoint(self, p, tol=tolerance.TOL):
        """Return the nearest _Point on the _Segment to the point p.

mapPoint(p [, tol])

The function has one required argument:

p: A _Point object or a tuple of two floats

There is a single optional argument:

tol: A float value equal or greater than 0.

This function is used to map a possibly near-by _Point object to an
actual _Point on the _Segment. If the distance between the actual
_Point and the _Point used as an argument is less than the Tolerance,
the actual _Point is returned. Otherwise, this function returns None.
        """
        _p = p
        if not isinstance(_p, point._Point):
            _p = point._Point(p)
        _t = tolerance.toltest(tol)
        _xp, _yp = _p.getCoords()
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _xmin = min(_x1, _x2) - _t - 1e-10
        _ymin = min(_y1, _y2) - _t - 1e-10
        _ap = None
        if _xp > _xmin and _yp > _ymin:
            _xmax = max(_x1, _x2) + _t + 1e-10
            _ymax = max(_y1, _y2) + _t + 1e-10
            if _xp < _xmax and _yp < _ymax:
                _sqlen = pow((_x2 - _x1), 2) + pow((_y2 - _y1), 2)
                _r = ((_xp - _x1)*(_x2 - _x1) + (_yp - _y1)*(_y2 - _y1))/_sqlen
                if _r < 0.0 : _r = 0.0
                if _r > 1.0 : _r = 1.0
                _px = _x1 + _r * (_x2 - _x1)
                _py = _y1 + _r * (_y2 - _y1)
                _xsep = abs(_px - _xp)
                _ysep = abs(_py - _yp)
                if ((_xsep < _t or abs(_xsep - _t) < 1e-10) and
                    (_ysep < _t or abs(_ysep - _t) < 1e-10)):
                    _ap = point._Point(_px, _py)
        return _ap

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not a _Segment exists within a region.

isRegion(xmin, ymin, xmax, ymax[, fully])

The four arguments define the boundary of an area, and the
method returns True if the _Segment lies within that area. If
the optional argument "fully" is used and is True, then both
endpoints of the _Segment must lie within the boundary.
Otherwise, the method returns False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        # print "xmin: %g" % _xmin
        # print "ymin: %g" % _ymin
        # print "xmax: %g" % _xmax
        # print "ymax: %g" % _ymax
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        # print "x1: %g" % _x1
        # print "y1: %g" % _y1
        # print "x2: %g" % _x2
        # print "y2: %g" % _y2
        _pxmin = min(_x1, _x2)
        _pymin = min(_y1, _y2)
        _pxmax = max(_x1, _x2)
        _pymax = max(_y1, _y2)
        # print "pxmin: %g" % _pxmin
        # print "pymin: %g" % _pymin
        # print "pxmax: %g" % _pxmax
        # print "pymax: %g" % _pymax
        _val = False
        if (_xmin < _pxmin and _xmax > _pxmax and
            _ymin < _pymin and _ymax > _pymax):
            _val = True
        elif (not fully and
              not (_pxmax < _xmin or _pymax < _ymin or
                   _pxmin > _xmax or _pymin > _ymax)):
            if _xmin < _x1 < _xmax and _ymin < _y1 < _ymax:
                _val = True
            elif _xmin < _x2 < _xmax and _ymin < _y2 < _ymax:
                _val = True
            else:
                #
                # the hard way ...
                #
                # see if segment intersects an imaginary segment
                # from (xmin, ymax) to (xmax, ymin)
                #
                # p1 = (xmin, ymax)
                # p2 = (xmax, ymin)
                # p3 = p1 of segment
                # p4 = p2 of segment
                #
                _d = ((_xmax - _xmin)*(_y2 - _y1)) - ((_ymin - _ymax)*(_x2 - _x1))
                if abs(_d) > 1e-10:
                    _n = ((_ymax - _y1)*(_x2 - _x1)) - ((_xmin - _x1)*(_y2 - _y1))
                    _r = _n/_d
                    if 0.0 < _r < 1.0:
                        _val = True
                if not _val:
                    #
                    # see if segment intersects an imaginary segment
                    # from (xmin, ymin) to (xmax, ymax)
                    #
                    # p1 = (xmin, ymin)
                    # p2 = (xmax, ymax)
                    # p3 = p1 of segment
                    # p4 = p2 of segment
                    #
                    _d = ((_xmax - _xmin)*(_y2 - _y1)) - ((_ymax - _ymin)*(_x2 - _x1))
                    if abs(_d) > 1e-10:
                        _n = ((_ymin - _y1)*(_x2 - _x1)) - ((_xmin - _x1)*(_y2 - _y1))
                        _r = _n/_d
                        if 0.0 < _r < 1.0:
                            _val = True

        return _val

    def clone(self):
        """Create an identical copy of a _Segment.

clone()  
        """
        _cp1 = self.__p1.clone()
        _cp2 = self.__p2.clone()
        return _Segment(_cp1, _cp2)

#
# Segment defaults
#

def_linetype = linetype.Linetype('Solid_Default', None)
def_color = color.Color(0xffffff)
def_thickness = 1.0

def_seg_style = style.Style("Segment_Default", def_linetype,
                            def_color, def_thickness)

class Segment(_Segment, tree.TreeItem, graphicobject.GraphicObject):
    """The storable and displayed representation of _Segment objects.

The Segment class is derived from _Segment, treeItem, and
GraphicObject classes, so it shares all the attributes and
methods of those classes.
    """
    def __init__(self, p1, p2, st=def_seg_style, lt=None, col=None, t=None):
        _p1 = p1
        if not isinstance(_p1, point.Point):
            _p1 = point.Point(p1)
        _p2 = p2
        if not isinstance(_p2, point.Point):
            _p2 = point.Point(p2)
        if _p1 == _p2:
            raise ValueError, "Segments must have two different endpoints."
        _Segment.__init__(self, _p1, _p2)
        tree.TreeItem.__init__(self)
        graphicobject.GraphicObject.__init__(self, st, lt, col, t)


    def isModified(self):
        """Indicate if the Segment has been changed.

isModified()

This method extends the TreeItem::isModified() method.
        """
        _p1, _p2 = self.getEndpoints()
        if _p1.isModified() or _p2.isModified():
            self.modified()
        return tree.TreeItem.isModified(self)
    
    def setP1(self, p):
        """Set the first endpoint Point of the Segment.

setP1(p)

This method extends _Segment::setP1() and sets the modified
flag indicating that the Segment has been altered.
        """
        _Segment.setP1(self, p)
        self.modified()
    
    p1 = property(_Segment.getP1, setP1, None,
                  "First endpoint of the Segment.")

    def setP2(self, p):
        """Set the second endpoint Point of the Segment.

setP2(p)

This method extends _Segment::setP2() and sets the modified
flag indicating that the Segment has been altered.
        """
        _Segment.setP2(self, p)
        self.modified()
    
    p2 = property(_Segment.getP2, setP2, None,
                  "Second endpoint of the Segment.")    

    def move(self, dx, dy):
        """Move a Segment.

move(dx, dy)

This method extends _Segment::move() and sets the modified
flag indicating that the Segment has been altered.
        """
        _Segment.move(self, dx, dy)
        self.modified()

    def clone(self):
        """Create an identical copy of a Segment.

clone()

This method overrides the _Segment::clone() method.
        """
        _p1, _p2 = self.getEndpoints()
        _cp1 = _p1.clone()
        _cp2 = _p2.clone()
        _st = self.getStyle()
        _lt = self.getLinetype()
        _col = self.getColor()
        _th = self.getThickness()
        return Segment(_cp1, _cp2, _st, _lt, _col, _th)

    def mapPoint(self, p, tol=tolerance.TOL):
        """Return the nearest Point on the Segment to the point p.

mapPoint(p [, tol])

This function extends _Segment::mapPoint().
        """
        _pt = _Segment.mapPoint(self, p, tol)
        if _pt is not None:
            _x, _y = _pt.getCoords()
            _pt = point.Point(_x, _y)
        return _pt

class SegmentTree(maptree.PointMapTree):
    """A Tree class specific for Segment objects.

This class is based on the PointMapTree class, and holds
only Segment objects.
    """
    def __init__(self):
        """Initialize a SegmentTree.

SegmentTree()
        """
        maptree.PointMapTree.__init__(self, Segment)

    def binscan(self, obj, tol=tolerance.TOL):
        """Find a Segment object held in a SegmentTree via binary search.

binscan(obj [, tol])

This method has one required parameter:

obj: Either a Segment object or a tuple of two-points

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if two
     Segments are close enough to be considered the same.

If a Segment object is found in the SegmentTree within the
tolerance, the index of that object in the SegmentTree is
returned. Otherwise, the function returns None.
        """
        _o = obj
        if isinstance(_o, Segment):
            _p1, _p2 = _o.getEndpoints()
        elif isinstance(_o, tuple):
            _l = len(_o)
            if _l != 2:
                raise ValueError, "Tuple must contain exactly two objects."
            _p1 = _o[0]
            if not isinstance(_p1, point.Point):
                _p1 = point.Point(_o[0])
            _p2 = _o[1]
            if not isinstance(_p2, point.Point):
                _p2 = point.Point(_o[1])
        else:
            raise TypeError, "Invalid type for SegmentTree search: " + `obj`
        _t = tolerance.toltest(tol)
        if self.isModified():
            raise ValueError, "SegmentTree in modified state."
        _ro = None
        if len(self):
            _xmin = min(_p1.x, _p2.x)
            _ymin = min(_p1.y, _p2.y)
            _xmax = max(_p1.x, _p2.x)
            _ymax = max(_p1.y, _p2.y)
            # print "xmin: %g" % _xmin
            # print "ymin: %g" % _ymin
            # print "xmax: %g" % _xmax
            # print "ymax: %g" % _ymax
            _scanlist = []
            _lo = 0
            _hi = len(self)
            # print "lo: 0; hi: %d" % _hi
            while _lo < _hi:
                _mid = (_hi+_lo)//2
                if _mid in _scanlist:
                    break
                _scanlist.append(_mid)
                # print "testing %d" % _mid
                _sp1, _sp2 = self[_mid].getEndpoints()
                # print _sp1
                # print _sp2
                _sxmin = min(_sp1.x, _sp2.x)
                # print "sxmin: %g" % _sxmin
                if (_xmin - _sxmin) > -1e-10:
                    _lo = _mid + 1
                else:
                    _hi = _mid
                    # print "lo: %d; hi: %d" % (_lo, _hi)
            for _i in xrange(_mid, -1, -1):
                # print "index: %d" % _i
                _seg = self[_i]
                _sp1, _sp2 = _seg.getEndpoints()
                _sxmin = min(_sp1.x, _sp2.x)
                if _sxmin < _xmin - _t - 1e-10:
                    break
                if abs(_sxmin - _xmin) < _t + 1e-10:
                    _symin = min(_sp1.y, _sp2.y)
                    if abs(_symin - _ymin) < _t + 1e-10:
                        _sxmax = max(_sp1.x, _sp2.x)
                        if abs(_sxmax - _xmax) < _t + 1e-10:
                            _symax = max(_sp1.y, _sp2.y)
                            if abs(_symax - _ymax) < _t + 1e-10:
                                _ro = _seg
                                break
        return _ro

    def scan(self, obj, tol=tolerance.TOL):
        """Find a Segment object held in a SegmentTree via linear search.

scan(obj [, tol])

This method has one required parameter:

obj: Either a Segment object or a tuple of two-points

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if two
     Segments are close enough to be considered equal.

If a Segment object is found in the SegmentTree within the
tolerance, the index of that object in the SegmentTree is
returned. Otherwise, the function returns None.
        """
        _o = obj
        if isinstance(_o, Segment):
            _p1, _p2 = _o.getEndpoints()
        elif isinstance(_o, tuple):
            _l = len(_o)
            if _l != 2:
                raise ValueError, "Tuple must contain exactly two objects."
            _p1 = _o[0]
            if not isinstance(_p1, point.Point):
                _p1 = point.Point(_o[0])
            _p2 = _o[1]
            if not isinstance(_p2, point.Point):
                _p2 = point.Point(_o[1])
        else:
            raise TypeError, "Invalid type for SegmentTree search: " + `obj`
        _t = tolerance.toltest(tol)
        if self.isModified():
            raise ValueError, "SegmentTree in modified state."
        _xmin = min(_p1.x, _p2.x)
        _ymin = min(_p1.y, _p2.y)
        _xmax = max(_p1.x, _p2.x)
        _ymax = max(_p1.y, _p2.y)
        _ro = None
        for _sobj in self:
            _sp1, _sp2 = _sobj.getEndpoints()
            _sxmin = min(_sp1.x, _sp2.x)
            if _sxmin > _xmin + _t + 1e-10:
                break
            if abs(_sxmin - _xmin) < _t + 1e-10:
                _symin = min(_sp1.y, _sp2.y)
                if abs(_symin - _ymin) < _t + 1e-10:
                    _sxmax = max(_sp1.x, _sp2.x)
                    if abs(_sxmax - _xmax) < _t + 1e-10:
                        _symax = max(_sp1.y, _sp2.y)
                        if abs(_symax - _ymax) < _t + 1e-10:
                            _ro = _sobj
                            break
        return _ro
