#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# tool stuff
#

import types
import array

class Tool(object):
    """A generic tool object.

This class is meant to be a base class for tools. A Tool
instance the following attributes:

list: A list the tool can use to store objects
handlers: A dictionary used to store functions

A Tool object has the following methods:

setHandler(): Set the handler for an event type.
getHandler(): Get the handler for an event type.
delHandler(): Remove the handler for an event type.
hasHandler(): See if a handler for an event type is stored.
clearHandlers(): Unset all the handlers in the tool.
initialize(): Retore the tool to its original state.
setLocation(): Store a location in the tool.
getLocation(): Get the last stored location in the tool.
    """
    def __init__(self):
        """Instantiate a Tool.

t = Tool()
        """
        self.__objlist = []
        self.__objtype = None
        self.__handlers = {}
        self.__location = None
        self.__curpoint = None
        self.__points = []
        self.__xpts = array.array('d')
        self.__ypts = array.array('d')
        self.__shift = None

    def __len__(self):
        """Return the number of objects in the list via len().
        """
        return len(self.__objlist)

    def __iter__(self):
        """Make the Tool iterable.
        """
        return iter(self.__objlist)

    def getList(self):
        """Return the Tool's object list.

getList()        
        """
        return self.__objlist

    list = property(getList, None, None, "Tool object list.")

    def initialize(self):
        """Initialize a Tool.

initialize()

This function purges the Tool's object list and handler dictionary.
        """
        del self.__objlist[:]
        self.__objtype = None
        self.__handlers.clear()
        self.__location = None
        self.__curpoint = None
        del self.__points[:]
        del self.__xpts[:]
        del self.__ypts[:]
        self.__shift = None

    def setHandler(self, key, func):
        """Set a handler for the Tool.

setHandler(key, func)

There are two arguments for this function:

key: A string used to identify a particular action
func: A function object

There are no restrictions on what the function 'func' does,
the argument count, etc. Any call to setHandler() with
a key that is already stored replaces the old 'func' argument
with the new one. The function argument may be None, and
the key argument must be a string.
        """
        if key is None:
            raise ValueError, "Key value cannot be None."
        if not isinstance(key, str):
            raise TypeError, "Invalid key: " + `key`
        if func and not isinstance(func, types.FunctionType):
            raise TypeError, "Invalid function: " + `func`
        self.__handlers[key] = func

    def getHandler(self, key):
        """Return the function for a particular key.

getHandler(key)

Given argument 'key', the function associated with it is
returned. A KeyError is raised if the argument 'key' had
not be used to store a function.
        """
        if not isinstance(key, str):
            raise TypeError, "Invalid argument to getHandler(): " + `key`
        if not self.__handlers.has_key(key):
            raise KeyError, "Invalid key '%s'" % key
        return self.__handlers[key]

    def delHandler(self, key):
        """Delete the handler associated with a particular key

delHandler(key)

The argument "key" should be a string.
        """
        if self.__handlers.has_key(key):
            del self.__handlers[key]
            
    def hasHandler(self, key):
        """Check if there is a handler stored for a given key.

hasHandler(key)

The 'key' argument must be a string. The function returns 1
if there is a handler for that key, 0 otherwise.
        """
        _k = key
        if not isinstance(_k, str):
            raise TypeError, "Invalid argument to hasHandler(): " + `key`
        return self.__handlers.has_key(_k)

    def clearHandlers(self):
        """Unset all the handlers for the tool.

clearHandlers()

This function does not alter the Tool's object list.
        """
        self.__handlers.clear()

    def setObjType(self, objtype):
        """Store the class of the object type being created.

setObjType(objtype)        
        """
        if not isinstance(objtype, types.TypeType):
            raise TypeError, "Objtype must be a type"
        self.__objtype = objtype

    def getObjType(self):
        """Return the type of object the tool will create.

getObjType()        
        """
        return self.__objtype
    
    def pushObject(self, obj):
        """Add an object to the Tool's object list.

pushObject(obj)        
        """
        self.__objlist.append(obj)

    def popObject(self):
        """Remove the last object on the Tool's object list.

popObject()

If the object list is empty, this function returns None.
        """
        _obj = None
        if len(self.__objlist):
            _obj = self.__objlist.pop()
        return _obj

    def delObjects(self):
        """Remove all objects from the Tool's object list.

delObjects()

This function does not alter the Tool's handlers.        
        """
        del self.__objlist[:]

    def getObject(self, idx):
        """Access an object in the tool.

getObject(idx)

The argument "idx" is the index into the list of
stored objects.
        """
        return self.__objlist[idx]
    
    def setLocation(self, x, y):
        """Store an x/y location in the tool.

setLocation(x,y)

Store an x-y coordinate in the tool. Both arguments
should be floats
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        self.__location = (_x,_y)

    def getLocation(self):
        """Return the stored location in the tool

getLocation()        
        """
        return self.__location

    def storeCoords(self, x, y):
        """Store an x/y location in the tool.

storeCoords(x,y)

Both the "x" and "y" arguments should be floats.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        self.__xpts.append(_x)
        self.__ypts.append(_y)

    def getCoordPair(self, idx):
        """Return a set of the stored x/y coordinates.

getCoordPair(idx)

The argument "idx" is the index of the stored coordinate
pair.
        """
        return (self.__xpts[idx], self.__ypts[idx])
    
    def getCoords(self):
        """Return the stored x/y coordinates.

getCoords()
        """
        _coords = []
        assert len(self.__xpts) == len(self.__ypts), "Coord array error"
        for _i in len(self.__xpts):
            _coords.append((self.__xpts[_i], self.__ypts[_i]))
        return _coords
    
    def clearLocation(self):
        """Reset the location to an empty value.

clearLocation()        
        """
        self.__location = None
        
    def setCurrentPoint(self, x, y):
        """Set the tool's current point.

setCurrentPoint(x,y)

Store an x-y coordinate in the tool. Both arguments
should be int.
        """
        _x = x
        if not isinstance(_x, int):
            _x = int(x)
        _y = y
        if not isinstance(_y, int):
            _y = int(y)
        self.__curpoint = (_x,_y)

    def getCurrentPoint(self):
        """Return the tool's current point value.

getCurrentPoint()        
        """
        return self.__curpoint

    def clearCurrentPoint(self):
        """Reset the current point to an empty value

clearCurrentPoint()        
        """
        self.__curpoint = None
        
    def setShift(self, x, y):
        """

setShift(x,y)

Store an "x" and "y" value in the tool representing
the amount to move objects. Both values should be floats.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        self.__shift = (_x,_y)

    def getShift(self):
        """Return the stored values for object movement.

getShift()
        """
        return self.__shift

    def setPoint(self, idx, x, y):
        """Save an x/y location in the points list at index idx

setPoint(idx, x, y)        
        """
        self.__points.insert(idx, (x,y))

    def getPoint(self, idx):
        """Get an x/y location stored in the points array from index idx.

getPoint(idx)
        """
        tup = self.__points[idx]
        return tup
        
class ToolCore(object):
    """A class representing the functionality of a Tool.

The ToolCore class provides a link between the Tool object
and what it can do at any given time.

A ToolCore instance has two attributes:

name: A string identifying the instance
init: A function called to initialize the tool

The function attribute is the most important, as it
sets the tool up to do whatever this core will have it
do. The actual function pointed to by the function
attribute should be of the form

def func(tool, *args)

Where 'tool' is the Tool object, and there are optional
arguments aftewards.
    """

    def __init__(self, name, init, image=None):
        """Instantiate a ToolCore object.

ToolCore(name, init [, image])

name: A string identifying this ToolCore object
init: A function that will initialize the tool

Neither name nor init can be None.

There is one optional argument:

image: A filename for an image to represent this object.
        """
        if not name or not isinstance(name, str):
            raise TypeError, "Invalid name for ToolCore: " + `name`
        if not init or not isinstance(init, types.FunctionType):
            raise TypeError, "Invalid function: " + `init`
        if image and not isinstance(image, str):
            raise TypeError, "Invalid image name: " + `image`
        self.__name = name
        self.__init = init
        self.__image = image

    def getName(self):
        """Return the name of the ToolCore object.

getName()        
        """
        return self.__name

    name = property(getName, None, None, "ToolCore name.")

    def getInitFunction(self):
        """Return the initializing function.

getInitFunction()        
        """
        return self.__init

    init = property(getInitFunction, None, None, "Initializing function.")

    def getImage(self):
        """Return the ToolCore's image name.

getImage()        
        """
        return self.__image

    def setImage(self, image):
        """Set the ToolCore's image.

setImage(image)        
        """
        if image and not isinstance(image, str):
            raise TypeError, "Invalide image name: " + `image`
        self.__image = image
        
class Toolset(object):
    """A class that contains common tool functionality.

The Toolset class is meant to be a base class for containing
a group of tool functions that are similar. For example, a
Toolset called "Polygons" could contain functionality for
triangles, rectangles, pentagons, etc.

A Toolset has the following attributes:

name: A string identifying the toolset
list: A list of the ToolCore objects in the toolset
    """
    def __init__(self, name):
        """Instantiate a Toolset.

Toolset(name)

The 'name' argument is a string that identifies the toolset.
        """
        if name is None:
            raise ValueError, "All Toolsets must have a name."
        if not isinstance(name, str):
            raise TypeError, "Invalid Toolset name: " + `name`
        self.__name = name
        self.__list = []

    def __len__(self):
        """Return the number of ToolCore objects in the Toolset via len().
        """
        return len(self.__list)
    
    def __iter__(self):
        """Make the Toolset iterable.
        """
        return iter(self.__list)

    def getName(self):
        """Return the Toolset's name.
        """
        return self.__name

    name = property(getName, None, None, "Toolset name.")

    def getList(self):
        """Return the Toolset's list.
        """
        return self.__list

    list = property(getList, None, None, None)

    def addToolCore(self, toolcore):
        """Add a ToolCore object to the ToolSet.

addToolCore(toolcore)

The toolcore argument must be a ToolCore object.
        """
        if not isinstance(toolcore, ToolCore):
            raise TypeError, "Invalid object: " + `toolcore`
        self.__list.append(toolcore)

class Toolbox(object):
    """A class representing a collection of Toolsets.

A Toolbox object contains toolsets. There is no limit
to the number of toolsets that a Toolbox can contain, though
the name of each Toolset must be unique.
    """
    def __init__(self):
        """Instantiate a Toolbox.

Toolbox()

There are no arguments needed to create a Toolbox object.
        """
        self.__sets = {}

    def __len__(self):
        """Return the number of Toolsets in the Toolbox via len().
        """
        return len(self.__sets)

    def __iter__(self):
        """Make the Toolbox iterable.

The iteration returns the Toolset objects.
        """
        return itervalues(self.__sets)
    
    def addToolset(self, toolset):
        """Add a Toolset to the Toolbox.

tbox = Toolbox()
tset = Toolset("setname")
tbox.addToolset(tset)

A toolset will be added to the toolbox if and only if there
is not already a toolset in the toolbox with the same name.
Trying to add a toolset with the same name raises an ValueError
        """
        if not isinstance(toolset, Toolset):
            raise TypeError, "Only Toolsets can be added to toolboxes: " + `toolset`
        name = toolset.getName()
        if self.__sets.has_key(name):
            raise ValueError, "Toolbox already has a Toolset '%s'" % name
        self.__sets[name] = toolset

    def delToolset(self, obj):
        """Remove a Toolset from the Toolbox.
        
There are two ways to invoke this method:

delToolset('name'): The argument is a string representing a Toolset name.
delToolset(toolset): The argument is a Toolset object.

If the toolset is not found in the Toolbox, a ValueError exception
is raised.
        """
        if isinstance(obj, str):
            _n = obj
        elif isinstance(obj, Toolset):
            _n = obj.getName()
        else:
            raise TypeError, "Invalid argument to hasToolset: " + `obj`
        if not self.__sets.has_key(_n):
            raise ValueError, "Invalid Toolset '%s'" % _n
        del self.__sets[_n]
        
    def hasToolset(self, obj):
        """Check if the Toolbox has a toolset of a particular name.

There are two ways to invoke this method:

hasToolset('name'): The argument is a string representing a Toolset name
hasToolset(toolset): The argument is a toolset itself

The function returns 1 if the toolset or toolset name is found, 0 otherwise.
        """
        if isinstance(obj, str):
            _n = obj
        elif isinstance(obj, Toolset):
            _n = obj.getName()
        else:
            raise TypeError, "Invalid argument to hasToolset: " + `obj`
        return self.__sets.has_key(_n)
